#pragma once
#include "stdafx.h"
#include <list>
#include <cassert>
#include <iostream>

#include "pkcs11stub.h"

using namespace std;

#define PKCS11_DLL "softhsm2.dll"

#define MAX_READER_NAME_LENGTH  256
#define TESTING_TOKEN_LABEL		 "pv204"
typedef unsigned long ULONG;
//typedef unsigned int DWORD;
typedef char TCHAR;
typedef void* CK_VOID_PTR;

#define TS_ASSERT(x) assert(x)
#define TS_FAIL(x) { cout << x; assert(false); }
//#define _T(x) x


typedef struct _ITEM_DATA_PKCS11READER {
	ULONG               slotID;
	ULONG               hSession;
	char                readerName[MAX_READER_NAME_LENGTH];
	bool                bCardPresent;
	char                cardATR[MAX_READER_NAME_LENGTH];

	_ITEM_DATA_PKCS11READER() {
		clear();
	}

	void clear() {
		slotID = 0;
		hSession = NULL;
		memset(readerName, 0, sizeof(readerName));
		bCardPresent = false;
		memset(cardATR, 0, sizeof(cardATR));
	}
} ITEM_DATA_PKCS11READER;

typedef list<ITEM_DATA_PKCS11READER> lp11r;

class PKCS11Test {
	CPKCS11Stub m_pkcs11Mngr;

	int getTestTokenSession(CK_SLOT_ID* pSlotID, CK_SLOT_INFO* pSlotInfo, CK_TOKEN_INFO* pTokenInfo) {
		int             status = 0;
		CK_SLOT_ID      pkcs11Slots[100];
		DWORD           pkcs11SlotsCount = 100;
		ITEM_DATA_PKCS11READER  readerInfo;
		m_pkcs11Mngr.Init(PKCS11_DLL);

		TS_ASSERT(m_pkcs11Mngr.C_GetSlotList(FALSE, pkcs11Slots, &pkcs11SlotsCount) == CKR_OK);
		TS_ASSERT(pkcs11SlotsCount > 0);
		bool	bTestTokenFound = FALSE;
		for (DWORD i = 0; i < pkcs11SlotsCount; i++) {
			TS_ASSERT(m_pkcs11Mngr.C_GetSlotInfo(pkcs11Slots[i], pSlotInfo) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_GetTokenInfo(pkcs11Slots[i], pTokenInfo) == CKR_OK);

			// IF MORE THAN 3 SPACES DETECTED, THAN CUT READER NAME
			TCHAR* pos = 0;
			if ((pos = strstr((TCHAR*)pTokenInfo->label, _T("   "))) != NULL) {
				memset(pTokenInfo->label + (pos - (TCHAR*)pTokenInfo->label) * sizeof(TCHAR), 0, sizeof(TCHAR));
			}

			// Do only when testing token was found
			if (strcmp((const char*)pTokenInfo->label, TESTING_TOKEN_LABEL) == 0) {
				bTestTokenFound = true;
				*pSlotID = pkcs11Slots[i];
				break;
			}
		}

		if (!bTestTokenFound) status = CKR_TOKEN_NOT_PRESENT;

		return status;
	}

public:
	void testListReaders(void) {
		cout << endl << "######### testListReaders #########" << endl;
		// PKCS#11 READERS
		int             status = 0;
		CK_SLOT_ID      pkcs11Slots[100];
		DWORD           pkcs11SlotsCount = 100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		ITEM_DATA_PKCS11READER  readerInfo;
		m_pkcs11Mngr.Init(PKCS11_DLL);

		if ((status = m_pkcs11Mngr.C_GetSlotList(FALSE, pkcs11Slots, &pkcs11SlotsCount)) == 0) {
			TS_ASSERT(pkcs11SlotsCount > 0);
			for (DWORD i = 0; i < pkcs11SlotsCount; i++) {
				if (m_pkcs11Mngr.C_GetSlotInfo(pkcs11Slots[i], &slotInfo) == 0) {
					if (m_pkcs11Mngr.C_GetTokenInfo(pkcs11Slots[i], &tokenInfo) == 0) {
						readerInfo.clear();
						readerInfo.slotID = pkcs11Slots[i];

						readerInfo.bCardPresent = true;
						tokenInfo.label[31] = 0x00;
						slotInfo.slotDescription[63] = 0x00;
						strcpy_s(readerInfo.cardATR, MAX_READER_NAME_LENGTH, (TCHAR*)tokenInfo.label);
						strcpy_s(readerInfo.readerName, MAX_READER_NAME_LENGTH, (TCHAR*)slotInfo.slotDescription);

						// IF MORE THAN 3 SPACES DETECTED, THAN CUT READER NAME
						TCHAR* pos = 0;
						if ((pos = strstr(readerInfo.cardATR, _T("   "))) != NULL) {
							memset(readerInfo.cardATR + (pos - readerInfo.cardATR) * sizeof(TCHAR), 0, sizeof(TCHAR));
						}
						if ((pos = strstr(readerInfo.readerName, _T("   "))) != NULL) {
							memset(readerInfo.readerName + (pos - readerInfo.readerName) * sizeof(TCHAR), 0, sizeof(TCHAR));
						}

						// Print slot&token info
						cout << "Slot info: ";
						cout << ": ";
						cout << readerInfo.readerName;
						cout << " / ";
						cout << readerInfo.cardATR << endl;

					}
					else  TS_FAIL("C_GetTokenInfo() fail");
				}
				else  TS_FAIL("C_GetSlotInfo fail");
			}
		}
		else  TS_FAIL("C_GetSlotList fail");
	}

	void testGetMechanismsList(void) {
		cout << endl << "######### testGetMechanismsList #########" << endl;

		int             status = 0;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {
			CK_ULONG numMechanisms = 0;
			TS_ASSERT(m_pkcs11Mngr.C_GetMechanismList(slotID, NULL_PTR, &numMechanisms) == CKR_OK);
			CK_MECHANISM_TYPE* mechanisms = new CK_MECHANISM_TYPE[numMechanisms];
			memset(mechanisms, 0, numMechanisms * sizeof(CK_MECHANISM_TYPE));
			TS_ASSERT(m_pkcs11Mngr.C_GetMechanismList(slotID, mechanisms, &numMechanisms) == CKR_OK);
			cout << "Supported mechanisms: " << endl;
			for (CK_ULONG i = 0; i < numMechanisms; i++) {
				cout << m_pkcs11Mngr.GetMechanimsName(mechanisms[i]);
				cout << " (0x";
				cout << hex << mechanisms[i];
				cout << ")" << endl;
			}
		}
		else  TS_FAIL("C_GetSlotList fail");
	}

	void testTokenPINTests(void) {
		cout << endl << "######### testTokenPINTests #########" << endl;
#ifdef DISABLE_SLOW_TESTS
		TS_FAIL("testTokenPINTests disabled");
#else
		// PKCS#11 READERS
		int             status = 0;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {
#define TEST_PIN	"1234"
#define TEST_PIN2	"123488"
#define DEFAULT_ADMIN_PIN			"123456"
#define WRONG_ADMIN_PIN				"123457"


			CK_SESSION_HANDLE   hSession;
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);

			cout << "Login user" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);
			cout << "Logging again should fail - but some apps are failing with that" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_USER_ALREADY_LOGGED_IN);
			TS_ASSERT(m_pkcs11Mngr.C_Logout(hSession) == CKR_OK);
			cout << "Wrong user PIN" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN2, strlen(TEST_PIN2)) == CKR_PIN_INCORRECT);

			cout << "Set User PIN after admin is logged in" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_SO, (CK_CHAR*)DEFAULT_ADMIN_PIN, strlen(DEFAULT_ADMIN_PIN)) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_InitPIN(hSession, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_Logout(hSession) == CKR_OK);

			TS_ASSERT(m_pkcs11Mngr.C_CloseAllSessions(slotID) == CKR_OK);
		}
		/**/
#endif
	}

	void testInsertObjects(void) {
		cout << endl << "######### testInsertObjects #########" << endl;

		// PKCS#11 READERS
		int             status = 0;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {

#define TEST_PIN	"1234"
#ifndef DEFAULT_ADMIN_PIN
#define DEFAULT_ADMIN_PIN			"123456"
#endif

#define NUM_OBJECTS		10
			CK_OBJECT_HANDLE hObject[NUM_OBJECTS];
			CK_ULONG ulObjectCount;
			// TEMPLATE to use for creation/searching object on token
			CK_CHAR label_private[] = { "Test1_private" };	//label of data object
			CK_OBJECT_CLASS dataClass = CKO_DATA;  //type of data in object
			CK_BBOOL ptrue = true; CK_BBOOL pfalse = FALSE;
			CK_CHAR	data_private[] = { "PV204 Private" };
			CK_ATTRIBUTE dataTemplate_private[] = {
				{ CKA_CLASS, &dataClass, sizeof(dataClass) },
				{ CKA_TOKEN, &ptrue, sizeof(ptrue) },
				{ CKA_LABEL, label_private, sizeof(label_private) },
				{ CKA_VALUE, (CK_VOID_PTR)data_private, sizeof(data_private) },
				{ CKA_PRIVATE, &ptrue, sizeof(ptrue) }  // private object
			};
			BYTE    numAttributes_private = 5;

			CK_CHAR label_public[] = { "Test1_public" };	//label of data object
			CK_CHAR	data_public[] = { "PV204 Public" };
			CK_ATTRIBUTE dataTemplate_public[] = {
				{ CKA_CLASS, &dataClass, sizeof(dataClass) },
				{ CKA_TOKEN, &ptrue, sizeof(ptrue) },
				{ CKA_LABEL, label_public, sizeof(label_public) },
				{ CKA_VALUE, (CK_VOID_PTR)data_public, sizeof(data_public) },
				{ CKA_PRIVATE, &pfalse, sizeof(pfalse) }  // private object
			};
			BYTE    numAttributes_public = 5;


			CK_SESSION_HANDLE	hSession;
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);

			cout << "Create new private object - should fail as we are not logged in" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_private, numAttributes_private, &hObject[0]) != CKR_OK);

			cout << "Create new public object - should success even when we are not logged in" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_public, numAttributes_public, &hObject[0]) == CKR_OK);

			// Login user
			TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);
			cout << "Create new object - should succedd" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_private, numAttributes_private, &hObject[0]) == CKR_OK);
			// Logout user
			TS_ASSERT(m_pkcs11Mngr.C_Logout(hSession) == CKR_OK);

			cout << "No private objects should be found as we are logged out and object is private" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_private, numAttributes_private) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK);
			TS_ASSERT(ulObjectCount == 0);

			cout << "At least one public object should be found even when we are logged out" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_public, numAttributes_public) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK);
			TS_ASSERT(ulObjectCount >= 1);

			cout << "Partial template searching - At least one public object should be found (CKA_CLASS, CKA_TOKEN, CKA_LABEL)" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_public, 3) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK);
			TS_ASSERT(ulObjectCount >= 1);

			cout << "Partial template searching - At least one public object should be found (CKA_CLASS, CKA_TOKEN)" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_public, 2) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK);
			TS_ASSERT(ulObjectCount >= 1);

			cout << "Login user" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_private, numAttributes_private) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK);
			cout << "	At least one private object should be present" << endl;
			TS_ASSERT(ulObjectCount >= 1);
			CK_ATTRIBUTE dataTemplate[] = { { CKA_VALUE, NULL_PTR, 0 }, };
			int numAttributes = 1;

			cout << "Partial template searching - At least oone public and one private object should be found (CKA_CLASS, CKA_TOKEN)" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_public, 2) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK);
			TS_ASSERT(ulObjectCount >= 2);

			cout << "Find private object" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_private, numAttributes_private) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK);
			TS_ASSERT(ulObjectCount >= 1);

			cout << "Load object value and check its content" << endl;
			TS_ASSERT(m_pkcs11Mngr.C_GetAttributeValue(hSession, hObject[0], dataTemplate, numAttributes) == CKR_OK);
			dataTemplate[0].pValue = new char[dataTemplate[0].ulValueLen + 1];
			memset(dataTemplate[0].pValue, 0, dataTemplate[0].ulValueLen + 1);
			TS_ASSERT(m_pkcs11Mngr.C_GetAttributeValue(hSession, hObject[0], dataTemplate, numAttributes) == CKR_OK);
			TS_ASSERT(dataTemplate[0].ulValueLen == sizeof(data_private));
			TS_ASSERT(memcmp(dataTemplate[0].pValue, data_private, sizeof(data_private)) == 0);
			string content = "	Object value: ";
			content.append((TCHAR*)dataTemplate[0].pValue);
			cout << content.c_str() << endl;
			delete[] dataTemplate[0].pValue;

			TS_ASSERT(m_pkcs11Mngr.C_DestroyObject(hSession, hObject[0]) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_CloseAllSessions(slotID) == CKR_OK);
			/**/
		}
	}

	void testRandom(void) {
		cout << endl << "######### testRandom #########" << endl;

		int             status = 0;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {
			CK_SESSION_HANDLE	hSession;
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);

			cout << "Small chunk of random data" << endl;
			BYTE	random1[10];
			TS_ASSERT(m_pkcs11Mngr.C_GenerateRandom(hSession, random1, sizeof(random1)) == CKR_OK);
			BYTE	random2[10];
			TS_ASSERT(m_pkcs11Mngr.C_GenerateRandom(hSession, random2, sizeof(random2)) == CKR_OK);
			DWORD	dwDiffer = 0;
			for (int i = 0; i < sizeof(random1); i++) if (random1[i] != random2[i]) dwDiffer++;
			TS_ASSERT(((float)dwDiffer / sizeof(random1)) > 0.9);		// 9 out of 10 must be different

			cout << "Big chunk of random data" << endl;
			BYTE	random3[1000];
			TS_ASSERT(m_pkcs11Mngr.C_GenerateRandom(hSession, random3, sizeof(random3)) == CKR_OK);
			BYTE	random4[1000];
			TS_ASSERT(m_pkcs11Mngr.C_GenerateRandom(hSession, random4, sizeof(random4)) == CKR_OK);
			dwDiffer = 0;
			for (int i = 0; i < sizeof(random3); i++) if (random3[i] != random4[i]) dwDiffer++;
			TS_ASSERT(((float)dwDiffer / sizeof(random3)) > 0.95);		// 90% of bytes must be different
		}
		TS_ASSERT(m_pkcs11Mngr.C_CloseAllSessions(slotID) == CKR_OK);
	}
};

